<?php class PlateImport extends Device {

    const USERNAME='plateimporter';
    const USERFULLNAME='Plate Importer';
    const PLATETYPE_DIMENSIONS_MISMATCH_MESSAGE='Plate type already exists (found by name) but dimensions do not match';
    const USER_BAD_EMAIL_MESSAGE='Email address is invalid';
    const SCORING_SYSTEM_NOT_FOUND_MESSAGE='No scoring system found with that name';
    const BAD_SEQUENCE_MESSAGE='Supplied sequence is not a valid DNA or protein sequence';
    const PROJECT_FOR_PROTEIN_ALREADY_EXISTS_MESSAGE = 'Project for this protein already exists';
    const ONLY_ONE_CONSTRUCT_PER_PLATE_MESSAGE = 'Multiple proteins/constructs per plate not supported';
    const NO_PLATE_WITH_BARCODE_MESSAGE = 'No plate with that barcode';
    const BAD_LIGHT_TYPE_MESSAGE = 'Only "UV" and "Visible" accepted as lightType';
    const BAD_DATE_MESSAGE = 'Bad date, should be a UTC date in YYYY-MM-DD hh:mm:ss';
    const NO_IMAGER_WITH_NAME_MESSAGE = 'No imager with that name';

    public static function getVersion(){
        return 1;
    }

    /**
     * @param $plateType
     * @return array|mixed
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
    public static function createPlateType($plateType){
        static::require('name',$plateType);
        static::require('rows', $plateType);
        static::require('columns', $plateType);
        static::require('subPositions', $plateType);
        $existingPlateType=platetype::getByName($plateType['name']);
        if($existingPlateType) {
            if($existingPlateType['rows']!=$plateType['rows'] || $existingPlateType['cols']!=$plateType['columns'] || $existingPlateType['subs']!=$plateType['subPositions']){
                throw new BadRequestException(PlateImport::PLATETYPE_DIMENSIONS_MISMATCH_MESSAGE);
            }
        } else {
            session::becomeAdmin();
            $sharedProject = project::getByName(project::SHARED);
            if (!$sharedProject) {
                throw new NotFoundException('No project called ' . project::SHARED);
            }
            $dropMapping = array('', '');
            for ($i = 1; $i <= $plateType['subPositions']; $i++) {
                $dropMapping[0] .= $i;
                $dropMapping[1] .= 'R';
            }
            $dropMapping = implode(',', $dropMapping);
            $existingPlateType = platetype::create(array(
                'name' => $plateType['name'],
                'projectid' => $sharedProject['id'],
                'rows' => $plateType['rows'],
                'cols' => $plateType['columns'],
                'subs' => $plateType['subPositions'],
                'dropmapping' => $dropMapping
            ));
            $existingPlateType = $existingPlateType['created'];
        }
        $plateType['dropMapping']=$existingPlateType['dropmapping'];
        $plateType['receiversId']=$existingPlateType['id'];
        return $plateType;
    }

    /**
     * @param $user
     * @return array|mixed
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
    public static function createUser($user){
        static::require('fullName', $user);
        static::require('email', $user);
        if(!validator::isValid($user['email'], validator::EMAIL)){
            throw new BadRequestException(static::USER_BAD_EMAIL_MESSAGE);
        }
        session::becomeAdmin();
        $existingUser=user::getByProperty('email', $user['email']);

        if($existingUser) {
            $existingUser=$existingUser['rows'][0];
        } else {
            $temporaryUsername = 'AUTO' . rand(10000000, 99999999);
            while (user::getByName($temporaryUsername)) {
                $temporaryUsername = 'AUTO' . rand(10000000, 99999999);
            }
            $existingUser = user::create(array(
                'fullname' => $user['fullName'],
                'email' => $user['email'],
                'name' => $temporaryUsername,
                'password' => 'ChangeMe!',
                'isactive' => 0
            ));
            $existingUser=$existingUser['created'];
        }
        session::revertAdmin();
        $user['receiversId']=$existingUser['id'];
        return $user;
    }

    /**
     * Creates an imager, or finds an existing one, and returns the supplied parameters with the imager ID in receiversId.
     * @param $imager
     * @throws BadRequestException if nominalTemperature is not numeric, or if plateCapacity is not an integer greater than 0.
     * @throws ServerException if imager already exists but temperature or capacity do not match
     * @throws NotFoundException
     * @throws ForbiddenException
     */
    public static function createImager($imager){
        static::require("name", $imager);
        static::require("manufacturerName", $imager);
        static::require("nominalTemperature", $imager);
        static::require("plateCapacity", $imager);
        if(!is_numeric($imager['nominalTemperature'])){
            throw new BadRequestException('Imager nominalTemperature must be numeric');
        }
        if(!is_numeric($imager['plateCapacity']) || !is_integer(1*$imager['plateCapacity'])){
            throw new BadRequestException('Imager plateCapacity must be an integer');
        } else if(1*$imager['plateCapacity']<=0){
            throw new BadRequestException('Imager plateCapacity must be at least 1');
        }
        $existingImager=imager::getByName($imager['name']);
        if($existingImager){
            if(1*$imager['nominalTemperature']!==1*$existingImager['temperature']){
                throw new ServerException('Imager already exists but temperatures do not match. Update either sender or IceBear.');
            } else if(1*$imager['plateCapacity']!==1*$existingImager['platecapacity']){
                throw new ServerException('Imager already exists but capacities do not match. Update either sender or IceBear.');
            }
        } else {
            session::becomeAdmin();
            $existingImager=imager::create(array(
                'name'=>$imager['name'],
                'friendlyname'=>$imager['name'],
                'manufacturer'=>$imager['manufacturerName'],
                'temperature'=>$imager['nominalTemperature'],
                'platecapacity'=>1*$imager['plateCapacity'],
                'alertlevel'=>(int)(0.75*$imager['plateCapacity']),
                'warninglevel'=>(int)(0.9*$imager['plateCapacity']),
            ))['created'];
            session::revertAdmin();
        }
        $imager['receiversId']=$existingImager['id'];
        return $imager;
    }

    /**
     * @param $params
     * @return array
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException|ForbiddenException
     */
    public static function createScoringSystem($params){
        static::require('name', $params);
        static::require('scores', $params);
        $existing=crystalscoringsystem::getByName($params['name']);
        if($existing){
            $params['receiversId']=$existing['id'];
            return $params;
        }
        session::becomeAdmin();
        $system=crystalscoringsystem::create(array(
            'name'=>$params['name']
        ))['created'];
        $system['scores']=[];
        foreach ($params['scores'] as $score){
            static::require('name', $score);
            static::require('index', $score);
            static::require('color', $score);
            static::require('hotkey', $score);
            crystalscore::create(array(
                'crystalscoringsystemid'=>$system['id'],
                'name'=>$params['name'].'_'.$score['name'],
                'label'=>$score['name'],
                'color'=>$score['color'],
                'scoreindex'=>$score['index'],
                'hotkey'=>$score['hotkey']
            ))['created'];
        }
        session::revertAdmin();
        return static::getScoringSystem(array('name'=>$system['name']));
    }

    /**
     * @param $params
     * @return array
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     */
    public static function getScoringSystem($params){
        static::require('name', $params);
        $existingScoringSystem=crystalscoringsystem::getByName($params['name']);
        if(!$existingScoringSystem){
            throw new NotFoundException(static::SCORING_SYSTEM_NOT_FOUND_MESSAGE);
        }
        $scores=crystalscoringsystem::getscores($existingScoringSystem['id']);
        if($scores['rows']){ $scores=$scores['rows']; }
        $params['scores']=$scores;
        $params['receiversId']=$existingScoringSystem['id'];
        return $params;
    }

    /**
     * @param $proteins
     * @param $projectOwnerId
     * @return mixed
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
    public static function createMultipleProteins($proteins, $projectOwnerId){
        foreach ($proteins as &$protein) {
            $protein=static::createProteinAndConstructs($protein, $projectOwnerId);
        }
        return $proteins;
    }

    /**
     * Finds the project and protein with a matching name, creating if necessary.
     * @return array $protein but with receiversId populated for protein and construct(s)
     * @param $protein array The supplied protein/construct/sequence hierarchy
     * @param $projectOwnerId int The owner of the project (only used if project does not exist)
     * @throws BadRequestException
     * @throws ServerException
     * @throws NotFoundException
     * @throws ForbiddenException
     */
    public static function createProteinAndConstructs($protein, $projectOwnerId){
        static::require('name', $protein);
        static::require('constructs', $protein);
        session::becomeAdmin();

        //try to find existing project, which will have our protein in it. If it exists, don't create.
        $proteinName=$protein['name'];
        try {
            $project=project::getByName($proteinName);
        } catch (NotFoundException $e){
            $project=project::create(array(
                'name'=>$proteinName,
                'description'=>'Generated by plate import',
                'owner'=>$projectOwnerId
            ))['created'];
            session::becomeAdmin();
        }
        $projectId=$project['id'];

        $existingProtein=protein::getByProperties(array(
            'name'=>$proteinName,
            'projectid'=>$projectId
        ));
        if($existingProtein){
            $existingProtein=$existingProtein['rows'][0];
        } else {
            $proteinAcronym='';
            if(isset($protein['proteinAcronym'])){
                $proteinAcronym=$protein['proteinAcronym'];
            }
            $existingProtein=protein::create(array(
                'name'=>$protein['name'],
                'description'=>$protein['name'],
                'projectid'=>$projectId,
                'proteinacronym'=>$proteinAcronym
            ))['created'];
            //"Full length" construct is created by default. Get rid of it.
            $unneededConstructs=protein::getconstructs($existingProtein['id']);
            if(!empty($unneededConstructs)){
                foreach ($unneededConstructs['rows'] as $unneededConstruct){
                    construct::delete($unneededConstruct['id']);
                }
            }
        }
        $proteinId=$existingProtein['id'];
        $protein['receiversId']=$proteinId;

        foreach($protein['constructs'] as &$construct) {
            $existingConstruct = construct::getByProperties(array(
                'name' => $construct['name'],
                'proteinid' => $proteinId
            ));
            if ($existingConstruct) {
                $existingConstruct = $existingConstruct['rows'][0];
            } else {
                $description = $construct['name'];
                if (isset($construct['description'])) {
                    $description = $construct['description'];
                }
                $existingConstruct = construct::create(array(
                    'name' => $construct['name'],
                    'description' => $description,
                    'owner' => $projectOwnerId,
                    'proteinid' => $proteinId
                ))['created'];
            }

            $constructId = $existingConstruct['id'];
            $construct['receiversId'] = $constructId;

            $existingSequences = construct::getsequences($constructId);
            if (isset($existingSequences['rows'])) {
                $existingSequences = $existingSequences['rows'];
            }
            $newSequences=[];
            if(isset($construct['sequences'])){
                $newSequences = $construct['sequences'];
            }

            if (empty($existingSequences) && empty($newSequences)) {
                $sequencesMatch = true;
            } else if (empty($existingSequences) && !empty($newSequences)) {
                $sequencesMatch = false;
            } else if (!empty($existingSequences) && empty($newSequences)) {
                $sequencesMatch = false;
            } else if (count($existingSequences) != count($newSequences)) {
                $sequencesMatch = false;
            } else {
                $sequencesMatch = true;
                for ($i = 0; $i < count($newSequences); $i++) {
                    $newSequence = $newSequences[$i];
                    $existingSequence = $existingSequences[$i];
                    $newSequence = preg_replace('/[^A-Z]/', '', strtoupper($newSequence));
                    $existingProteinSequence = preg_replace('/[^A-Z]/', '', strtoupper($existingSequence['proteinsequence']));
                    $existingDnaSequence = preg_replace('/[^A-Z]/', '', strtoupper($existingSequence['dnasequence']));
                    if ($newSequence !== $existingDnaSequence && $newSequence !== $existingProteinSequence) {
                        $sequencesMatch = false;
                        break;
                    }
                }
            }

            if (!$sequencesMatch) {
                //First delete the existing sequences
                if (!empty($existingSequences)) {
                    foreach ($existingSequences as $existingSequence) {
                        sequence::delete($existingSequence['id']);
                    }
                }
                //Now create with the new information
                if (!empty($newSequences)) {
                    $count = 1;
                    foreach ($newSequences as $newSequence) {
                        $toCreate = array(
                            'name' => 'Sequence ' . $count,
                            'constructid' => $constructId
                        );
                        if (validator::isValid($newSequence, validator::DNASEQUENCE)) {
                            $toCreate['dnasequence'] = $newSequence;
                            $toCreate['proteinsequence'] = sequence::dnaToProtein($newSequence);
                        } else if (validator::isValid($newSequence, validator::PROTEINSEQUENCE)) {
                            $toCreate['dnasequence'] = '';
                            $toCreate['proteinsequence'] = $newSequence;
                        } else {
                            throw new BadRequestException(static::BAD_SEQUENCE_MESSAGE);
                        }
                        sequence::create($toCreate);
                        $count++;
                    }
                }
            }
        }
        session::revertAdmin();
        return $protein;
    }

    /**
     * @param $plate
     * @return mixed
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     * @throws ForbiddenException
     * @throws NotModifiedException
     */
    public static function createPlate($plate){
        static::require('barcode',$plate);
        static::require('scoringSystemName',$plate);
        static::require('plateType',$plate);
        static::require('owner',$plate);
        session::becomeAdmin();

        $barcode=$plate['barcode'];
        $existingPlate=plate::getByName($barcode);
        if($existingPlate){
            $plate['receiversId']=$existingPlate['id'];
            $plate['plateType']=PlateImport::createPlateType($plate['plateType']);
            $plate['owner']=PlateImport::createUser($plate['owner']);
        } else {
            $scoringSystem=PlateImport::getScoringSystem(array('name'=>$plate['scoringSystemName']));
            if(!$scoringSystem){
                session::revertAdmin();
                throw new NotFoundException('No scoring system called '.$plate['scoringSystemName'].'. Create it in IceBear.');
            }
            $projectId=static::getDefaultProjectId();

            $plate['plateType']=PlateImport::createPlateType($plate['plateType']);
            $plate['owner']=PlateImport::createUser($plate['owner']);

            $newPlate=plate::create(array(
                'name'=>$barcode,
                'ownerid'=>$plate['owner']['receiversId'],
                'projectid'=>$projectId,
                'platetypeid'=>$plate['plateType']['receiversId'],
                'crystalscoringsystemid'=>$scoringSystem['receiversId']
            ))['created'];

            $plate['receiversId']=$newPlate['id'];
        }

        if(isset($plate['proteins']) && !empty($plate['proteins'])){
            if(count($plate['proteins'])!==1 || count($plate['proteins'][0]['constructs'])!==1){
                throw new ServerException(self::ONLY_ONE_CONSTRUCT_PER_PLATE_MESSAGE);
            }
            $plate['proteins']=PlateImport::createMultipleProteins($plate['proteins'], $plate['owner']['receiversId']);
            plate::update($plate['receiversId'], array(
                'constructid'=>$plate['proteins'][0]['constructs'][0]['receiversId']
            ));
        }
        session::revertAdmin();

        if(isset($plate['imageBatches'])){
            foreach ($plate['imageBatches'] as &$imagingSession){
                $imagingSession['plateBarcode']=$barcode;
                $imagingSession=PlateImport::createImageBatch($imagingSession);
                unset($imagingSession['plateBarcode']);
            }
        }

        return $plate;
    }

    /**
     * @param $imageBatch
     * @return array
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     * @throws ForbiddenException
     */
    public static function createImageBatch($imageBatch){
        static::require('dateTimeUtc', $imageBatch);
        static::require('plateBarcode',$imageBatch);
        static::require('lightType',$imageBatch);
        static::require('imagerName',$imageBatch);
        try {
            session::becomeAdmin();
            $dateTime=$imageBatch['dateTimeUtc'];
            $dateTime=DateTime::createFromFormat('Y-m-d H:i:s', $dateTime, new DateTimeZone('UTC'));
            if(!$dateTime){
                throw new BadRequestException(self::BAD_DATE_MESSAGE);
            }

            //Ensure the imager already exists
            $imagerName=$imageBatch['imagerName'];
            $imager=imager::getByName($imagerName);
            if(!$imager) {
                throw new NotFoundException(self::NO_IMAGER_WITH_NAME_MESSAGE);
            }
            $imagerId=$imager['id'];

            //Ensure the plate already exists
            $barcode=$imageBatch['plateBarcode'];
            $plate=plate::getByName($barcode);
            if(!$plate) {
                throw new NotFoundException(self::NO_PLATE_WITH_BARCODE_MESSAGE);
            }
            $plateId=$plate['id'];

            //subtract 1 second from imaged time for UV images, to separate from visible
            $lightType=$imageBatch['lightType'];
            if(!in_array($lightType, array('UV', 'Visible'))){
                throw new BadRequestException(self::BAD_LIGHT_TYPE_MESSAGE);
            }
            if('UV'===$lightType){
                $dateTime->sub(new DateInterval('PT1S'));
            }

            $existingImagingSession=imagingsession::getByProperties(array(
                'imageddatetime'=>$dateTime->format('Y-m-d H:i:s'),
                'plateid'=>$plateId,
                'lighttype'=>$lightType
            ));
            if($existingImagingSession){
                $existingImagingSession=$existingImagingSession['rows'][0];
            } else {
                $sendersId=0;
                if(isset($imageBatch['sendersId'])){
                    $sendersId=$imageBatch['sendersId'];
                }

                $imagingParametersName=$imagerName.'_'.$lightType;
                $imagingParametersVersionName=$imagingParametersName.'_v1';
                $imagingParametersVersion=imagingparametersversion::getByName($imagingParametersVersionName);
                if(!$imagingParametersVersion){
                    $sharedProjectid=static::getSharedProjectId();
                    $imagingParameters=imagingparameters::getByName($imagingParametersName);
                    if(!$imagingParameters){
                        $imagingParameters=imagingparameters::create(array(
                            'name'=>$imagingParametersName,
                            'manufacturer'=>$imager['manufacturer'],
                            'manufacturerdatabaseid'=>0,
                            'projectid'=>$sharedProjectid
                        ))['created'];
                    }
                    $imagingParametersVersion=imagingparametersversion::create(array(
                        'name'=>$imagingParametersVersionName,
                        'projectid'=>$sharedProjectid,
                        'manufacturerdatabaseid'=>0,
                        'imagingparametersid'=>$imagingParameters['id']
                    ))['created'];
                }
                $imagingParametersVersionId=$imagingParametersVersion['id'];
                $existingImagingSession=imagingsession::create(array(
                    'name'=>$barcode.'_'.$lightType.'_'.$dateTime->format('Y-m-d H:i:s'),
                    'plateid'=>$plateId,
                    'lighttype'=>$lightType,
                    'imageddatetime'=>$dateTime->format('Y-m-d H:i:s'),
                    'imagerid'=>$imagerId,
                    'manufacturerdatabaseid'=>$sendersId,
                    'imagingparametersversionid'=>$imagingParametersVersionId
                ))['created'];
            }

            $imageBatch['receiversId']=$existingImagingSession['id'];

            if(isset($imageBatch['images'])){
                foreach ($imageBatch['images'] as &$image) {
                    $image['imagingSessionId']=$imageBatch['receiversId'];
                    $image=PlateImport::createImage($image);
                    unset($image['imagingSessionId']);
                    $image['imageData']='(trimmed)';
                    $image['thumbData']='(trimmed)';
                }
            }

        } finally {
            session::revertAdmin();
        }
        return $imageBatch;
    }

    /**
     * @param $image
     * @throws BadRequestException
     * @throws NotFoundException
     * @throws ServerException
     * @throws ForbiddenException
     */
    public static function createImage($image){
        static::require('rowNumber',$image);
        static::require('columnNumber',$image);
        static::require('subPositionNumber',$image);
        static::require('imageData', $image);
        static::require('thumbData', $image);
        $micronsPerPixel=-1;
        if(isset($image['micronsPerPixel'])){
            $micronsPerPixel=1*$image['micronsPerPixel'];
            if(0>=$micronsPerPixel){
                throw new BadRequestException('micronsPerPixel must be numeric and greater than zero');
            }
        }
        $row=1*$image['rowNumber'];
        $col=1*$image['columnNumber'];
        $sub=1*$image['subPositionNumber'];
        if(!$row){
            $rowLabels=platetype::$rowLabels;
            $row=array_search($image['rowNumber'], $rowLabels);
        }
        session::becomeAdmin();
        if(isset($image['imagingSessionId'])){
            $imagingSession=imagingsession::getById($image['imagingSessionId']);
            if(!$imagingSession){
                throw new NotFoundException('No imaging session with ID '.$image['imagingSessionId']);
            }
            $plate=plate::getById($imagingSession['plateid']);
            $barcode=$plate['name'];
        } else {
            static::require('plateBarcode', $image);
            static::require('dateTimeUtc', $image);
            static::require('imagerName', $image);
            static::require('lightType', $image);
            $barcode=$image['plateBarcode'];
            $plate=plate::getByName($barcode);
            if(!$plate){
                throw new NotFoundException('No plate found with barcode '.$image['barcode']);
            }
            $imager=imager::getByName($image['imagerName']);
            if(!$imager){
                throw new NotFoundException('No imager found with name '.$image['imagerName']);
            }
            $imagingSession=imagingsession::getByProperties(array(
                'plateid'=>$plate['id'],
                'imagerid'=>$imager['id'],
                'imageddatetime'=>$image['dateTimeUtc'],
                'lighttype'=>$image['lightType']
            ));
            if(!$imagingSession){
                throw new NotFoundException('No matching imaging session found for image');
            }
            $imagingSession=$imagingSession['rows'][0];
        }
        $imagingSessionId=$imagingSession['id'];
        $projectId=$imagingSession['projectid'];

        $plateWell=platewell::getByProperties(array(
            'plateid'=>$plate['id'],
            'row'=>$row,
            'col'=>$col
        ));
        if(!$plateWell){
            throw new BadRequestException('Bad plate row/column number');
        }
        $plateWell=$plateWell['rows'][0];
        $wellDrop=welldrop::getByProperties(array(
            'platewellid'=>$plateWell['id'],
            'dropnumber'=>$sub
        ));
        if(!$wellDrop){
            throw new BadRequestException('Bad plate subposition number');
        }
        $wellDrop=$wellDrop['rows'][0];

        $imageStore=rtrim(config::get('core_imagestore'),'/').'/';
        $imagePath=substr($barcode,0,2).'/'.substr($barcode,0,3).'/'.$barcode.'/';
        $imagePath=$imageStore.$imagePath.'imagingsession'.$imagingSessionId.'/';
        $thumbPath=$imagePath.'thumbs/';

        if(!file_exists($thumbPath) && !@mkdir($thumbPath, 0755, true)){
            throw new ServerException('Could not create directory for image storage');
        }

        $imageName=platewell::getWellLabel($row, $col).'_'.$sub.'.tmp';
        $imagePath.=$imageName;
        $thumbPath.=$imageName;

        if(!file_put_contents($imagePath, convert_uudecode($image['imageData'])) ){
            throw new ServerException('Could not write image to image store');
        }
        if(!@file_put_contents($thumbPath, convert_uudecode($image['thumbData'])) ){
            throw new ServerException('Could not write thumbnail to image store');
        }
        if(strlen(@file_get_contents($imagePath))<10){
            //probably a 2-byte garbage file caused by UUencode header/footer
            $imageBytes=static::stripUUencodingHeaderAndFooterLines($image['imageData']);
            $imageBytes=convert_uudecode($imageBytes);
            if(!@file_put_contents($imagePath, $imageBytes) ){
                throw new ServerException('Could not write image to image store');
            }
        }
        if(strlen(@file_get_contents($thumbPath))<10){
            //probably a 2-byte garbage file caused by UUencode header/footer
            $thumbBytes=static::stripUUencodingHeaderAndFooterLines($image['thumbData']);
            $thumbBytes=convert_uudecode($thumbBytes);
            if(!@file_put_contents($thumbPath, $thumbBytes) ){
                throw new ServerException('Could not write thumbnail to image store');
            }
        }

        $imageProperties=@getimagesize($imagePath);
        $thumbProperties=@getimagesize($thumbPath);
        if(!$imageProperties || !$imageProperties[0] || !$thumbProperties || !$thumbProperties[0]){
            throw new ServerException('Could not determine size of image and/or thumbnail. Is it corrupt?');
        }
        if(IMAGETYPE_JPEG==$imageProperties[2]){
            $newImagePath=substr($imagePath, 0, -4).'.jpg';
        } else if(IMAGETYPE_PNG==$imageProperties[2]){
            $newImagePath=substr($imagePath, 0, -4).'.png';
        } else {
            throw new ServerException('Image must be PNG or JPEG');
        }
        if(IMAGETYPE_JPEG==$thumbProperties[2]){
            $newThumbPath=substr($thumbPath, 0, -4).'.jpg';
        } else if(IMAGETYPE_PNG==$thumbProperties[2]){
            $newThumbPath=substr($thumbPath, 0, -4).'.png';
        } else {
            throw new ServerException('Thumbnail must be PNG or JPEG');
        }
        if(!@rename($imagePath, $newImagePath) || !@rename($thumbPath, $newThumbPath)){
            throw new ServerException('Could not save image/thumb in final store');
        }
        $width=$imageProperties[0];
        $height=$imageProperties[1];

        $existing=dropimage::getByProperties(array(
            'imagingsessionid'=>$imagingSessionId,
            'welldropid'=>$wellDrop['id']
        ));
        if(empty($existing)){
            $created=dropimage::create(array(
                'name'=>$barcode.'_'.platewell::getWellLabel($row, $col).'.'.$sub.'_is'.$imagingSessionId,
                'pixelheight'=>$height,
                'pixelwidth'=>$width,
                'micronsperpixelx'=>$micronsPerPixel,
                'micronsperpixely'=>$micronsPerPixel,
                'projectid'=>$projectId,
                'welldropid'=>$wellDrop['id'],
                'imagingsessionid'=>$imagingSessionId,
                'imagestorepath'=>$imageStore,
                'imagepath'=>str_replace($imageStore,'',$newImagePath),
                'thumbnailpath'=>str_replace($imageStore,'',$newThumbPath),
            ))['created'];
        } else {
            $existingImageId=$existing['rows'][0]['id'];
            $created=dropimage::update($existingImageId, array(
                'pixelheight'=>$height,
                'pixelwidth'=>$width,
                'micronsperpixelx'=>$micronsPerPixel,
                'micronsperpixely'=>$micronsPerPixel,
            ))['updated'];
        }

        $image['receiversId']=$created['id'];

        if(isset($image['marks'])){
            foreach($image['marks'] as &$mark){
                $mark=static::createCrystalMark($mark, $image['receiversId']);
            }
        }

        return $image;
    }

    private static function createCrystalMark($crystalMark, $dropImageId){
        $dropImage=dropimage::getById($dropImageId);
        if(!$dropImage){
            throw new NotFoundException("No drop image with ID $dropImageId");
        }
        if(isset($crystalMark['x'])){
            $x=(int)$crystalMark['x'];
        } elseif(isset($crystalMark['pixelX'])){
            $x=(int)$crystalMark['pixelX'];
        } else{
            throw new BadRequestException('Crystal pixelX is required');
        }
        if(isset($crystalMark['y'])){
            $y=(int)$crystalMark['y'];
        } elseif(isset($crystalMark['pixelY'])){
            $y=(int)$crystalMark['pixelY'];
        } else{
            throw new BadRequestException('Crystal pixelY is required');
        }
        if(!isset($crystalMark['type']) || "point"!==$crystalMark['type']){
            throw new BadRequestException('Crystal mark type is required, and must be "point"');
        }
        $crystal=crystal::getByProperties(array(
            'dropimageid'=>$dropImageId,
            'pixelx'=>$x,
            'pixely'=>$y
        ));
        if(!empty($crystal)){
            $crystal=$crystal['rows'][0];
        } else if(isset($crystalMark['name'])){
            $crystal=crystal::getByName($crystalMark['name']);
            if(!empty($crystal)){
                crystal::update($crystal['id'],array(
                    'pixelx'=>$x,
                    'pixely'=>$y
                ));
            }
        }

        if(empty($crystal)){
            $numberInDrop=1;
            $crystals=crystal::getByProperties(array(
                'dropimageid'=>$dropImageId
            ));
            if(!empty($crystals)){
                $numberInDrop+=1*$crystals['total'];
            }
            $crystal=crystal::create(array(
                'dropimageid'=>$dropImageId,
                'welldropid'=>$dropImage['welldropid'],
                'projectid'=>$dropImage['projectid'],
                'numberindrop'=>$numberInDrop,
                'name'=>'dummy'.time(),
                'prefix'=>'dummy'.time(),
                'pixelx'=>$x,
                'pixely'=>$y
            ))['created'];
        }

        if(isset($crystalMark['name'])){
            crystal::update($crystal['id'],array(
                'name'=>$crystalMark['name']
            ));
        }

        $crystalMark['receiversId']=$crystal['id'];
        return $crystalMark;
    }

    private static function stripUUencodingHeaderAndFooterLines($rawString){
        $parts=explode("\n", $rawString, 2);
        $headerlessString=$parts[1];
        $parts=explode("\nend", $headerlessString);
        return $parts[0];
    }


}
